
# # RSA Encryption

# In this document, we talk about the neccessary tools to perform RSA encryption.  Below we have the functions PadMessage and DepadMessage, we are thoroughly commented if you wish to read over how these functions work.  What these functions do is turn our String message into a list of numbers, each of which is no bigger than our pulic modulus.  Once again, If you have not done so, I would highly recommend watching Dr. Misseldine's [RSA Encryption](https://www.youtube.com/watch?v=LcKquwXkhzU&list=PLz7t89zv8Lp2D6xQOG7kUEbN1KP5u-mpH&index=79) video, he explains the math behind RSA encryption, and this code was adapted from his program written in MAGMA.  

# In[ ]:
import math


def PadMessage(message: str, N):
    # create the array that we will use to save our list of numbers
    PW = []
    # create a new string that we will add on to to create each number
    digits = "" 
    # for each letter in the message, encode it and add that encoding
    # onto the string digits.
    for letter in message:
        if letter== "a": digits = digits + "10"; 
        if letter== "b": digits = digits + "11"; 
        if letter== "c": digits = digits + "12";
        if letter== "d": digits = digits + "13"; 
        if letter== "e": digits = digits + "14"; 
        if letter== "f": digits = digits + "15"; 
        if letter== "g": digits = digits + "16"; 
        if letter== "h": digits = digits + "17"; 
        if letter== "i": digits = digits + "18"; 
        if letter== "j": digits = digits + "19"; 
        if letter== "k": digits = digits + "20"; 
        if letter== "l": digits = digits + "21";
        if letter== "m": digits = digits + "22"; 
        if letter== "n": digits = digits + "23"; 
        if letter== "o": digits = digits + "24"; 
        if letter== "p": digits = digits + "25"; 
        if letter== "q": digits = digits + "26"; 
        if letter== "r": digits = digits + "27"; 
        if letter== "s": digits = digits + "28"; 
        if letter== "t": digits = digits + "29"; 
        if letter== "u": digits = digits + "30"; 
        if letter== "v": digits = digits + "31"; 
        if letter== "w": digits = digits + "32"; 
        if letter== "x": digits = digits + "33"; 
        if letter== "y": digits = digits + "34"; 
        if letter== "z": digits = digits + "35"; 
        if letter== "A": digits = digits + "36"; 
        if letter== "B": digits = digits + "37"; 
        if letter== "C": digits = digits + "38"; 
        if letter== "D": digits = digits + "39"; 
        if letter== "E": digits = digits + "40"; 
        if letter== "F": digits = digits + "41"; 
        if letter== "G": digits = digits + "42"; 
        if letter== "H": digits = digits + "43"; 
        if letter== "I": digits = digits + "44"; 
        if letter== "J": digits = digits + "45"; 
        if letter== "K": digits = digits + "46"; 
        if letter== "L": digits = digits + "47"; 
        if letter== "M": digits = digits + "48"; 
        if letter== "N": digits = digits + "49"; 
        if letter== "O": digits = digits + "50"; 
        if letter== "P": digits = digits + "51";
        if letter== "Q": digits = digits + "52"; 
        if letter== "R": digits = digits + "53"; 
        if letter== "S": digits = digits + "54"; 
        if letter== "T": digits = digits + "55"; 
        if letter== "U": digits = digits + "56"; 
        if letter== "V": digits = digits + "57"; 
        if letter== "W": digits = digits + "58"; 
        if letter== "X": digits = digits + "59"; 
        if letter== "Y": digits = digits + "60"; 
        if letter== "Z": digits = digits + "61"; 
        if letter== " ": digits = digits + "62"; 
        if letter== ".": digits = digits + "63"; 
        if letter== ",": digits = digits + "64"; 
        if letter== "?": digits = digits + "65"; 
        if letter== "!": digits = digits + "66";
        if letter== "$": digits = digits + "67"; 
        if letter== "%": digits = digits + "68"; 
        if letter== "&": digits = digits + "69";
        if letter== "0": digits = digits + "70"; 
        if letter== "1": digits = digits + "71"; 
        if letter== "2": digits = digits + "72"; 
        if letter== "3": digits = digits + "73"; 
        if letter== "4": digits = digits + "74"; 
        if letter== "5": digits = digits + "75"; 
        if letter== "6": digits = digits + "76"; 
        if letter== "7": digits = digits + "77"; 
        if letter== "8": digits = digits + "78"; 
        if letter== "9": digits = digits + "79"; 
        if letter== "+": digits = digits + "80"; 
        if letter== "-": digits = digits + "81"; 
        if letter== "*": digits = digits + "82"; 
        if letter== "/": digits = digits + "83"; 
        if letter== "=": digits = digits + "84"; 
        if letter== "^": digits = digits + "85"; 
        if letter== "@": digits = digits + "86"; 
        if letter== "#": digits = digits + "87"; 
        if letter== "(": digits = digits + "88";
        if letter== ")": digits = digits + "89";
        if letter== ";": digits = digits + "90"; 
        if letter== ":": digits = digits + "91"; 
        if letter== "<": digits = digits + "92"; 
        if letter== ">": digits = digits + "93"; 
        if letter== "'": digits = digits + "94"; 
        if letter== "[": digits = digits + "95"; 
        if letter== "]": digits = digits + "96"; 
        if letter== "{": digits = digits + "97"; 
        if letter== "}": digits = digits + "98"; 
        if letter== "_": digits = digits + "99";
        
        # if the digit gets too big, that is bigger than N, then chop off the 
        # last two digits and add that number into the array, then continue the process
        if int(digits) > N:
            PW.append(int(digits[0:len(digits)-2]))
            digits = digits[len(digits)-2:]
    # at the end, if there are any digits that have not been added into the array, add them in
    if len(digits)!=0:
        PW.append(int(digits))
    # return the array of numbers
    return PW


    


# In[ ]:



def DepadMessage(digits):
    # define a some new strings for us to use
    letters = ""
    plaintext = ""
    # add all the digits in the array to the string letters
    for i in digits:
        letters = letters + str(i)
    
    # go through each pair of numbers in the string letters and add their corresponding plaintext to the 
    # plaintext string.
    for i in range(0,len(letters)+1,2):
        if letters[i:i+2]== "10": plaintext = plaintext + "a"; 
        if letters[i:i+2]== "11": plaintext = plaintext + "b"; 
        if letters[i:i+2]== "12": plaintext = plaintext + "c"; 
        if letters[i:i+2]== "13": plaintext = plaintext + "d";
        if letters[i:i+2]== "14": plaintext = plaintext + "e"; 
        if letters[i:i+2]== "15": plaintext = plaintext + "f"; 
        if letters[i:i+2]== "16": plaintext = plaintext + "g"; 
        if letters[i:i+2]== "17": plaintext = plaintext + "h"; 
        if letters[i:i+2]== "18": plaintext = plaintext + "i";
        if letters[i:i+2]== "19": plaintext = plaintext + "j";
        if letters[i:i+2]== "20": plaintext = plaintext + "k"; 
        if letters[i:i+2]== "21": plaintext = plaintext + "l"; 
        if letters[i:i+2]== "22": plaintext = plaintext + "m";
        if letters[i:i+2]== "23": plaintext = plaintext + "n"; 
        if letters[i:i+2]== "24": plaintext = plaintext + "o";
        if letters[i:i+2]== "25": plaintext = plaintext + "p"; 
        if letters[i:i+2]== "26": plaintext = plaintext + "q"; 
        if letters[i:i+2]== "27": plaintext = plaintext + "r"; 
        if letters[i:i+2]== "28": plaintext = plaintext + "s";
        if letters[i:i+2]== "29": plaintext = plaintext + "t"; 
        if letters[i:i+2]== "30": plaintext = plaintext + "u"; 
        if letters[i:i+2]== "31": plaintext = plaintext + "v"; 
        if letters[i:i+2]== "32": plaintext = plaintext + "w"; 
        if letters[i:i+2]== "33": plaintext = plaintext + "v"; 
        if letters[i:i+2]== "34": plaintext = plaintext + "y"; 
        if letters[i:i+2]== "35": plaintext = plaintext + "z"; 
        if letters[i:i+2]== "36": plaintext = plaintext + "A"; 
        if letters[i:i+2]== "37": plaintext = plaintext + "B"; 
        if letters[i:i+2]== "38": plaintext = plaintext + "C"; 
        if letters[i:i+2]== "39": plaintext = plaintext + "D"; 
        if letters[i:i+2]== "40": plaintext = plaintext + "E"; 
        if letters[i:i+2]== "41": plaintext = plaintext + "F"; 
        if letters[i:i+2]== "42": plaintext = plaintext + "G"; 
        if letters[i:i+2]== "43": plaintext = plaintext + "H"; 
        if letters[i:i+2]== "44": plaintext = plaintext + "I"; 
        if letters[i:i+2]== "45": plaintext = plaintext + "J"; 
        if letters[i:i+2]== "46": plaintext = plaintext + "K"; 
        if letters[i:i+2]== "47": plaintext = plaintext + "L"; 
        if letters[i:i+2]== "48": plaintext = plaintext + "M"; 
        if letters[i:i+2]== "49": plaintext = plaintext + "N"; 
        if letters[i:i+2]== "50": plaintext = plaintext + "O";
        if letters[i:i+2]== "51": plaintext = plaintext + "P";
        if letters[i:i+2]== "52": plaintext = plaintext + "Q"; 
        if letters[i:i+2]== "53": plaintext = plaintext + "R"; 
        if letters[i:i+2]== "54": plaintext = plaintext + "S"; 
        if letters[i:i+2]== "55": plaintext = plaintext + "T"; 
        if letters[i:i+2]== "56": plaintext = plaintext + "U"; 
        if letters[i:i+2]== "57": plaintext = plaintext + "V"; 
        if letters[i:i+2]== "58": plaintext = plaintext + "W"; 
        if letters[i:i+2]== "59": plaintext = plaintext + "X"; 
        if letters[i:i+2]== "60": plaintext = plaintext + "Y"; 
        if letters[i:i+2]== "61": plaintext = plaintext + "Z"; 
        if letters[i:i+2]== "62": plaintext = plaintext + " "; 
        if letters[i:i+2]== "63": plaintext = plaintext + "."; 
        if letters[i:i+2]== "64": plaintext = plaintext + ","; 
        if letters[i:i+2]== "65": plaintext = plaintext + "?"; 
        if letters[i:i+2]== "66": plaintext = plaintext + "!"; 
        if letters[i:i+2]== "67": plaintext = plaintext + "$"; 
        if letters[i:i+2]== "68": plaintext = plaintext + "%"; 
        if letters[i:i+2]== "69": plaintext = plaintext + "&";
        if letters[i:i+2]== "70": plaintext = plaintext + "0"; 
        if letters[i:i+2]== "71": plaintext = plaintext + "1"; 
        if letters[i:i+2]== "72": plaintext = plaintext + "2"; 
        if letters[i:i+2]== "73": plaintext = plaintext + "3"; 
        if letters[i:i+2]== "74": plaintext = plaintext + "4"; 
        if letters[i:i+2]== "75": plaintext = plaintext + "5"; 
        if letters[i:i+2]== "76": plaintext = plaintext + "6"; 
        if letters[i:i+2]== "77": plaintext = plaintext + "7"; 
        if letters[i:i+2]== "78": plaintext = plaintext + "8"; 
        if letters[i:i+2]== "79": plaintext = plaintext + "9"; 
        if letters[i:i+2]== "80": plaintext = plaintext + "+"; 
        if letters[i:i+2]== "81": plaintext = plaintext + "-"; 
        if letters[i:i+2]== "82": plaintext = plaintext + "*"; 
        if letters[i:i+2]== "83": plaintext = plaintext + "/"; 
        if letters[i:i+2]== "84": plaintext = plaintext + "="; 
        if letters[i:i+2]== "85": plaintext = plaintext + "^"; 
        if letters[i:i+2]== "86": plaintext = plaintext + "@"; 
        if letters[i:i+2]== "87": plaintext = plaintext + "#"; 
        if letters[i:i+2]== "88": plaintext = plaintext + "("; 
        if letters[i:i+2]== "89": plaintext = plaintext + ")";
        if letters[i:i+2]== "90": plaintext = plaintext + ";";
        if letters[i:i+2]== "91": plaintext = plaintext + ":"; 
        if letters[i:i+2]== "92": plaintext = plaintext + "<"; 
        if letters[i:i+2]== "93": plaintext = plaintext + ">"; 
        if letters[i:i+2]== "94": plaintext = plaintext + "'"; 
        if letters[i:i+2]== "95": plaintext = plaintext + "["; 
        if letters[i:i+2]== "96": plaintext = plaintext + "]"; 
        if letters[i:i+2]== "97": plaintext = plaintext + "{"; 
        if letters[i:i+2]== "98": plaintext = plaintext + "}"; 
        if letters[i:i+2]== "99": plaintext = plaintext + "_";
    # after that is done, return the plaintext string
    return plaintext




def to_binary(b: int):
    # if there is a negative number, return an error
    if b < 0:
        return "Please enter positive numbers"
    # otherwise, continue to compute the number in binary
    else:
        # define a new list that will hold all the digits of b in binary
        binary = []
        # define a new variable to represent the quotient, and set
        # it to one so we actually enter the next while loop
        q = 1
        # while the quotient is nonzero...
        while q!=0:
            # find the quotient of b/2
            q = math.floor(b/2)
            # as well as the remainder
            r = b % 2
            # take the remainder to be the digit of b in
            # binary that we need
            binary.append(r)
            # then set b to be the quotient
            b = q
        # return the list that contains the digits of b in binary
        return binary





def bin_exp(a: int, b: int, n: int):
    # test to ensure all values are positive.
    if a<0 or b<0 or n<0:
        return "Please enter positive integers only."
    else:
        # create a new list to hold b in binary
        binary = []
        # define quotient to be 1 so we enter the while loop
        q = 1
        # compute b in binary and save the result in reverse in binary list
        while q!=0:
            q = math.floor(b/2)
            r = b % 2
            binary.append(r)
            b = q
        # define a new list, that currently contains one value: a mod n
        squares = [a % n]
        # up to the length of b in binary, append a^2, a^4, a^8, ... onto the 
        # end of the list, this is done by taking the previous value and squaring
        # it, then reducing mod n
        for i in range(1,len(binary)):
            squares.append(squares[i-1]**2 % n)
        # define a variable to hold the final result
        result = 1
        # go through the list of squares, and if the corresponding binary digit is 1, 
        # then multiply it into the result, and then reduce mod n.
        for i in range(0,len(squares)):
            if binary[i]==1:
                result = (result * squares[i]) % n
        # return the final result
        return result




def ExpRSA(plaintext, e, N):
    # define a new list that we will store each encrypted (or decrypted) number in
    ciphertext = []
    
    # for each 'word' in the plaintext, raise it to the power of e mod n, then store
    # it to the ciphertext list.
    for word in plaintext:
        ciphertext.append(bin_exp(word, e, N))
    
    # return the ciphertext list
    return ciphertext

def EncryptRSA(plaintext, e, N):
    # first pad the plaintext into numbers, then encrypt it useing ExpRSA.
    return ExpRSA(PadMessage(plaintext, N), e, N)

def DecryptRSA(ciphertext, d, N):
    # first decrypt the message, then depad it into plaintext to read.
    return DepadMessage(ExpRSA(ciphertext, d, N))








def conjugacy_class(element , group):
    # define a set to hold all the conjugates of the element
    conjugates = set(())
    # for each item in the group, calculate the conjugate and add
    # it to the set defined in the line above.
    for i in group.list():
        conjugates.add((i**(-1))*element*i)
    # return the final set of all conjugates of the element.
    return conjugates


def all_conjugacy_classes(group):
    # define a list that will hold all the conjugacy classes of the group.
    conjugacy_classes = []
    # for each element in the group, 
    for i in group.list():
        # define a new set that will be the conjugacy class of that element
        conjugates = set(())
        # for each element of the group, add j^{-1}*i*j (remember that Sage 
        # takes this from right to left) to the conjugacy class
        for j in group.list():
            conjugates.add((j**(-1))*i*j)
        # if that conjugate class is not already in the list, add it.
        if conjugates not in conjugacy_classes:
            conjugacy_classes.append(conjugates)
    
    # return the final list of conjugacy classes.
    return conjugacy_classes


def conjugacy_class_gens(group):
    # define the dictionary that will hold all the conjugacy classes,
    # as well as which elements generate those classes.
    conjugacy_classes = {}
    # for each element in the group, 
    for i in group.list():
        # create a new set, that will be the conjugacy class of that element.
        conjugates = set(())
        for j in group.list():
            conjugates.add((j**(-1))*i*j)
        # sort that conjugacy class
        conjugates = sorted(conjugates)
        # if the conjugacy class is already found among the keys of the 
        # dictionary defined above, add the element to it's list, if not,
        # create a new entry with that conjugacy class.
        if str(conjugates) not in conjugacy_classes.keys():
            conjugacy_classes[str(conjugates)] = [i]
        else:
            conjugacy_classes[str(conjugates)].append(i)
    # return the final dictionary with all conjugacy classes and generators.
    return conjugacy_classes





#!/usr/bin/env python
# coding: utf-8

# # Algebraic Coding Theory

# SageMath provides a great environment to play with the concepts of Algebraic Coding Theory.  If you have not already, I would
# highly recommend watching Dr. Andrew Misseldine's 
# [Abstract Algebra](https://www.youtube.com/watch?v=Ox-_-D5bMyk&list=PLz7t89zv8Lp2D6xQOG7kUEbN1KP5u-mpH&index=80) videos on
# Coding Theory and Binary Symmetric Channels, note that the link leads to the first of several videos on this topic.  
# 
# In SageMath, there are a few new topics we need to be introduced to before we can experiment with Group Codes and encoding 
# messages and transmitting them over a binary symmetric channel.  The first is a new class object called a `MatrixSpace`.  
# Creating a MatrixSpace will allow us to create matrices and utilize them as our encoding and check matrices for our code. 
# The second is the command `GF(2)`, which specifies a Field of size 2.  We will use this field for our Matrix Space, since 
# `GF(2)` is implemented as the set {0,1} in SageMath.  
# 
# In the code below, we specify two Matrix Spaces, both over `GF(2)`, one of which consists of all 6x3 matrices and the other 
# all 3x6 matrices.  This will give us the tools to define our canonical parity check matrices.  We also define G and H just 
# as Dr. Misseldine defines them in the example he uses in his videos.  

# In[1]:




# Here we define yet another Matrix Space, again over `GF(2)`, this time consisting of all 3x1 matrices, this will be the 
# set of all messages we can encode.  we also define one specific message for us to use.  

# In[2]:




# Here we define some new classes for our use in this code.  We define the Bernoulli class, which acts as a Bernoulli 
# distribution.  Note that objects of type Bernoulli require a parameter `p`, which will be the probability of success. 
# We also define the class SymmetricChannel.  This will act as a Binary Symmetric Channel for us to send messages over 
# with some probability of errors.  Each object of type SymmetricChannel needs a Bernoulli distribution to accompany it.
# To test whether or not the SymmetricChannel is set up correctly with a Bernoulli distribution, I have provided a function 
# `test()`, which will tell us whether or not our SymmetricChannel is ready to go.  Do not change anything in this cell.

# In[1]:


# import the random module to use in the Bernoulli class
import random

# define the Bernoulli class
class Bernoulli:
    # define the constructor, that requires a parameter p, which should
    # be between zero and one.
    def __init__(self, p):
        self.p = p
    
    # define the method that will display information about the 
    # object when in a print statement
    def __repr__(self):
        return f"Bernoulli distribution with parameter {self.p}"
    
    # define the method that performs a random trial from the 
    # Bernoulli distribution based on the parameter p
    def randomTrial(self):
        value = random.random()
        if value < self.p:
            return 1
        else:
            return 0

# define the SymmetricChannel class
class SymmetricChannel:
    
    # define the constructor, which takes in a bernoulli object
    def __init__(self, b):
        # if b is a Bernoulli assign it to the bernoulli instance in the class,
        # if not, set the bernoulli instance to None
        if type(b) is Bernoulli:
            self.bernoulli = b
        else:
            self.bernoulli = None
            
    # define the method that will provide information about the object when
    # in a print statement
    def __repr__(self):
        return f"A SymmetricChannel based on a Bernoulli distribution with parameter {self.bernoulli.p}"
    
    # define a method to test whether or not the SymmetricChannel has been set up correctly.
    # this just means whether or not the bernoulli object was accepted correctly.
    def test(self):
        if type(self.bernoulli) is None:
            return "Not set up correctly"
        if type(self.bernoulli) is Bernoulli:
            return "Set up correctly"
        
    # this method takes in a matrix, which should be over GF(2), and uses the 
    # bernoulli object to randomly assign errors to the message, to simulate
    # errors we would see over a real channel.
    def transmit(self, message):
        # is the matrix is of the right type, continue to use the Binary Symmetric Channel
        if type(message) is sage.matrix.matrix_mod2_dense.Matrix_mod2_dense:
            transmitted = []
            # for each of the numbers in the message
            for i in range(0,len(message.rows())):
                # take a random trial from the bernoulli distribution.
                value = self.bernoulli.randomTrial()
                # then based on that random trial, either leave the message alone,
                # or commit an error in that bit.
                if value==0:
                    if message[i] == 0:
                        transmitted.append(0)
                    else:
                        transmitted.append(1)
                else:
                    if message[i] == 0:
                        transmitted.append(1)
                    else:
                        transmitted.append(0)
            # return the new message.
            return matrix(transmitted).transpose()
        # if the message was not of the right type, just return it.
        else:
            return matrix(message)
        


# Here is our first example of sending a message over the channel, we define a Bernoulli distribution, a SymmetricChannel 
# over that Bernoulli distribution, then we test the channel to ensure that it is set up correctly.  Then in the following
# cell, we define and encode the message, using our encoding message G, then transmit it over the symmetric channel, we 
# can then perform the check with our matrix H, to see if any errors occured in transmitting.

# In[4]:






def left_coset(H, g):
    H_elements = H.list()
    coset = set(())
    for element in H.list():
        coset.add(element*g)
        
    return sorted(coset)

def right_coset(H, g):
    H_elements  = H.list()
    coset = set(())
    for element in H_elements:
        coset.add(g*element)
        
    return sorted(coset)
 


def coset_generators(H, G):
    cosets  = {}
    for element in G.list():
        
        left_coset = set(())
        right_coset = set(())
        for item in H.list():
            right_coset.add(element*item)
            left_coset.add(item*element)
               
        right_coset = sorted(right_coset)
        left_coset = sorted(left_coset)
        
        
        if str(left_coset) not in cosets.keys() and str(right_coset) not in cosets.keys():
            cosets[str(left_coset)] = [str(element)+"H"]
            cosets[str(right_coset)] = ["H"+str(element)]
        if str(right_coset) in cosets.keys() and str("H"+str(element)) not in cosets[str(right_coset)]:
            cosets[str(right_coset)].append("H"+str(element))
            pass
        if str(left_coset) in cosets.keys() and str(str(element)+"H") not in cosets[str(left_coset)]:
            cosets[str(left_coset)].append(str(element)+"H")
    return cosets







def gcd_full(a: int,b: int):
    # store the original values of a and b into the 
    # variables aOrig and bOrig, to be used at the end
    aOrig = a
    bOrig = b
    
    # let atemp be the maximum of the absolute values of a and b
    # let btemp be the minimum
    atemp = max(abs(a),abs(b))
    btemp = min(abs(a),abs(b))
    
    # reassign the max value to be a and the minimum value to be b
    a = atemp
    b = btemp
    
    # create a list to hold all of the quotients, there is no need to 
    # save all the remainders, as they are not used in the pseudocode 
    # above
    quotients = []
    
    # same as before, go through and perform the steps of the Euclidean
    # algorithm, only this time, save all of the quotients into the list
    # that we just defined.
    while b > 0:
        quotient = math.floor(a/b)
        remainder = a % b
        a = b
        b = remainder
        quotients.append(quotient)
    
    # let d, be the value of the gcd
    d = a
    
    # throw the very last equation (the one with no remainder)
    # away, since we do not use it in the pseudocode above
    quotients.pop()
    
    # set x and y to their initial values
    x = 1
    y = -quotients.pop()
    
    # set the count equal to one, this will be used to alternate which one of x and y
    # we change in each step
    count = 0
    
    # follow the pseudocode above until there are no longer any quotients left
    while len(quotients)!=0:
        if count % 2 == 0:
            x = x-quotients.pop()*y
        if count % 2 == 1:
            y = y-quotients.pop()*x
        
        count = count + 1
    
    # this part is a little messy (sorry about that), but test to see which linear 
    # combination of the original values of a and b give us the gcd, then
    # return those values
    if x*aOrig+y*bOrig==d:
        return [d,x,y]
    elif (-x)*aOrig+y*bOrig==d:
        return [d,-x,y]
    elif x*aOrig+(-y)*bOrig==d:
        return [d,x,-y]
    elif (-x)*aOrig+(-y)*bOrig==d:
        return [d,-x,-y]
    elif y*aOrig+x*bOrig==d:
        return [d,y,x]
    elif (-y)*aOrig+x*bOrig==d:
        return [d,-y,x]
    elif y*aOrig+(-x)*bOrig==d:
        return [d,y,-x]
    elif (-y)*aOrig+(-x)*bOrig==d:
        return [d,-y,-x]








def inverse_mod(a, b):
    nums = gcd_full(a,b)
    if (a*nums[1])%b == 0:
        return nums[1]%b
    else:
        return nums[2]%b





def digitize(string):
    cipher_text = []
    for i in string:
        if i == 'A' or i == 'a': cipher_text.append(0)
        if i == 'B' or i == 'b': cipher_text.append(1)
        if i == 'C' or i == 'c': cipher_text.append(2)
        if i == 'D' or i == 'd': cipher_text.append(3)
        if i == 'E' or i == 'e': cipher_text.append(4)
        if i == 'F' or i == 'f': cipher_text.append(5)
        if i == 'G' or i == 'g': cipher_text.append(6)
        if i == 'H' or i == 'h': cipher_text.append(7)
        if i == 'I' or i == 'i': cipher_text.append(8)
        if i == 'J' or i == 'j': cipher_text.append(9)
        if i == 'K' or i == 'k': cipher_text.append(10)
        if i == 'L' or i == 'l': cipher_text.append(11)
        if i == 'M' or i == 'm': cipher_text.append(12)
        if i == 'N' or i == 'n': cipher_text.append(13)
        if i == 'O' or i == 'o': cipher_text.append(14)
        if i == 'P' or i == 'p': cipher_text.append(15)
        if i == 'Q' or i == 'q': cipher_text.append(16)
        if i == 'R' or i == 'r': cipher_text.append(17)
        if i == 'S' or i == 's': cipher_text.append(18)
        if i == 'T' or i == 't': cipher_text.append(19)
        if i == 'U' or i == 'u': cipher_text.append(20)
        if i == 'V' or i == 'v': cipher_text.append(21)
        if i == 'W' or i == 'w': cipher_text.append(22)
        if i == 'X' or i == 'x': cipher_text.append(23)
        if i == 'Y' or i == 'y': cipher_text.append(24)
        if i == 'Z' or i == 'z': cipher_text.append(25)
    return cipher_text
        


def alphabetize(digits):
    plain_text = ""
    for i in digits:
        if i == 0: plain_text = plain_text + "A"
        if i == 1: plain_text = plain_text + "B"
        if i == 2: plain_text = plain_text + "C"
        if i == 3: plain_text = plain_text + "D"
        if i == 4: plain_text = plain_text + "E"
        if i == 5: plain_text = plain_text + "F"
        if i == 6: plain_text = plain_text + "G"
        if i == 7: plain_text = plain_text + "H"
        if i == 8: plain_text = plain_text + "I"
        if i == 9: plain_text = plain_text + "J"
        if i == 10: plain_text = plain_text + "K"
        if i == 11: plain_text = plain_text + "L"
        if i == 12: plain_text = plain_text + "M"
        if i == 13: plain_text = plain_text + "N"
        if i == 14: plain_text = plain_text + "O"
        if i == 15: plain_text = plain_text + "P"
        if i == 16: plain_text = plain_text + "Q"
        if i == 17: plain_text = plain_text + "R"
        if i == 18: plain_text = plain_text + "S"
        if i == 19: plain_text = plain_text + "T"
        if i == 20: plain_text = plain_text + "U"
        if i == 21: plain_text = plain_text + "V"
        if i == 22: plain_text = plain_text + "W"
        if i == 23: plain_text = plain_text + "X"
        if i == 24: plain_text = plain_text + "Y"
        if i == 25: plain_text = plain_text + "Z"
    return plain_text
        



def CeasarEncrypt(plain_text, a, b):
    message = digitize(plain_text)
    for i in range(0,len(message)):
        message[i] = (a*message[i]+b) % 26
    return alphabetize(message)


def CeasarDecrypt(cipher_text, a, b):
    cipher_text = digitize(cipher_text)
    inverse = inverse_mod(a, 26)
    for i in range(0,len(cipher_text)):
        cipher_text[i] = (inverse*cipher_text[i]-inverse*b) % 26
    return alphabetize(cipher_text)
    






#!/usr/bin/env python

alphabet = {'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z'}
alpha = {'a':0, 'b':1,'c':2,'d':3,'e':4,'f':5,
         'g':6,'h':7,'i':8,'j':9,'k':10,'l':11,
         'm':12,'n':13,'o':14,'p':15,'q':16,'r':17,
         's':18,'t':19,'u':20,'v':21,'w':22,'x':23,'y':24,'z':25}



letter_frequencies = {'a': 0.08028354673082419,
 'h': 0.024342984615100982,
 'e': 0.1177437939585004,
 'd': 0.04119170745646784,
 'i': 0.07500530036964317,
 'n': 0.05773807877731995,
 'g': 0.029569609984974606,
 's': 0.09322197948065596,
 'l': 0.055759888646147324,
 'r': 0.07212189928375874,
 'v': 0.009651281767650231,
 'k': 0.014186554575371256,
 'w': 0.011983444410644986,
 'o': 0.060955172699870025,
 'f': 0.015259533751832085,
 'b': 0.022700330927426418,
 'c': 0.0359890489754155,
 'u': 0.03669883760589217,
 't': 0.05739885512015708,
 'm': 0.028445009817206383,
 'p': 0.029337315524091332,
 'y': 0.018312546666297946,
 'x': 0.0033184922983324574,
 'j': 0.0025957984200289446,
 'z': 0.004387784261128472,
 'q': 0.0018012038752615617}


def create_frequencies(filename):
    letters = {}
    count = 0
    with open(filename, 'r') as file:
        for word in file:
            for letter in word:
                if letter in alphabet:
                    if letter in letters.keys():
                        letters[letter.lower()] = letters[letter.lower()]+1
                    else:
                        letters[letter.lower()] = 1
                    count = count+1
    for k in letters.keys():
        letters[k] = float(letters[k]/count)
    return letters





def analize(ciphertext: str, filename):
    # test the input to ensure that the ciphertext was a string.
    if type(ciphertext)!=str:
        raise Exception("Incorrect type passed to analize function.")
    else:
        # define the dictionary that we will use to store the frequencies.
        frequencies = {}
        count = 0
        # go through each letter in the ciphertext.
        for letter in ciphertext.lower():
            # if that letter is in the alphabet,
            if letter in alphabet:
                # either add it to the frequencies dictionary, or if the key already exists,
                # increment its count.
                if letter in frequencies.keys():
                    frequencies[letter] = frequencies[letter]+1
                else:
                    frequencies[letter] = 1
                count = count + 1
        # go through and divide each count by the total.
        for k in frequencies.keys():
            frequencies[k] = float(frequencies[k]/count)
        # get the frequencies from the file passed into the method.
        common = create_frequencies(filename)
        # sort both the ciphertext frequencies and the dictionaries frequencies.
        frequencies_sorted = sorted(((value,key) for (key,value) in frequencies.items()),reverse=True)
        common_sorted = sorted(((value,key) for (key,value) in common.items()),reverse=True)
        most_likely = []
        # get the top five most likely options for the shift cipher.
        for i in range(0,5):
            if frequencies_sorted[0][0]>0.16:
                pos = 1
            else:
                pos = 0
            a = common_sorted[i][1]
            b = frequencies_sorted[pos][1]
            c = alpha[b]-alpha[a]
            most_likely.append((a,b,c % 26))
        # return the most likely options.
        return most_likely




